#include "game/EnemyEntity.hpp"

#include <cmath>
#include <SFML/Graphics/Sprite.hpp>
#include <SFML/Graphics/Texture.hpp>
#include <util/Logger.hpp>
#include <util/Math.hpp>

#include "game/Player.hpp"
#include "game/Room.hpp"

EnemyEntity::EnemyEntity( Room& room, Player& player, const std::string& theType, sf::Vector2i pos )
   : Entity::Entity( room, player ),
     type( theType ),

     stunTime( 0 ),
     startPos( pos ),

     minionTimer( 0 ),
     leafTimer( 0 ),
     bombTimer( 0 )
{
	spr.setPosition( sf::Vector2f( pos ) );
	spr.setScale( 0.25f, 0.25f );
}

void EnemyEntity::update()
{
	if ( health <= 0 )
	{
		return;
	}

	if ( type == "dungeonBoss" or type == "towerBoss" )
	{
		if ( ++minionTimer >= 650 and rand() % ( ( type == "dungeonBoss" ) ? 300 : 200 ) == 0 )
		{
			minionTimer = 0;
			player.playSound( "res/sound/spawn.wav" );
			for ( int i = 0; i < 8; ++i )
			{
				int x = 5 + ( rand() % 20 );
				int y = 5 + ( rand() % 20 );
				sf::Vector2f pos( x, y );

				if ( rand() % 2 != 0 or util::distance( pos, player.getPosition() ) < 5.f )
				{
					continue;
				}

				std::string monster = ( type == "dungeonBoss" ) ? "dungeonMonster" : "towerMonster";

				room.spawnEnemy( monster, pos );
			}
		}
	}

	float diffX = spr.getPosition().x - player.getPosition().x;
	float diffY = spr.getPosition().y - player.getPosition().y;
	float dirRad = std::atan2( diffX, diffY );
	float dirDeg = util::radiansToDegrees( dirRad );

	if ( stunTime >= 0 )
	{
		--stunTime;
		spr.rotate( 360 / 35.f );
		return;
	}
	else
	{
		spr.setRotation( -dirDeg );
	}


	if ( getCollisionRect().intersects( player.getCollisionRect() ) )
	{
		player.hurt( strength );
	}

	if ( type != "towerBoss" )
	{
		float offX = std::sin( dirRad ) * -speed;
		float offY = std::cos( dirRad ) * -speed;
		spr.move( offX, offY );

		Entity::update();
	}
	else
	{
		if ( ++leafTimer >= 100 and rand() % 150 == 0 )
		{
			leafTimer = 0;
			player.playSound( "res/sound/blow.wav" );

			float diffX = player.getPosition().x - spr.getPosition().x;
			float diffY = player.getPosition().y - spr.getPosition().y;
			float dirRad = std::atan2( diffX, diffY );

			player.push( 0.75f + ( ( rand() % 11 ) / 10.f * 0.25f ), dirRad );

			for ( std::vector< std::shared_ptr< Entity > >::iterator it = room.getEntities().begin(); it != room.getEntities().end(); ++it )
			{
				Entity* entity = it->get();

				if (  !dynamic_cast< EnemyEntity* >( entity ) )
				{
					float diffX = entity->getPosition().x - spr.getPosition().x;
					float diffY = entity->getPosition().y - spr.getPosition().y;
					float dirRad = std::atan2( diffX, diffY );

					entity->push( 0.5f, dirRad );
				}
			}
		}
		if ( ++bombTimer >= 200 and rand() % 250 == 0 )
		{
			bombTimer = 0;
			player.playSound( "res/sound/drop.wav" );
			for ( std::size_t i = 0; i < 9; ++i )
			{
				/*
				if ( i == 4 )
				{
					c
				//*/

				int x = 2 + ( rand() % 26 );//( ( i % 3 ) * 3 ) - 3;
				int y = 2 + ( rand() % 26 );//( ( i / 3 ) * 3 ) - 3;
				sf::Vector2f pos( /*player.getPosition().x +*/ x, /*player.getPosition().y +*/ y );

				room.makeBombAt( pos );
			}
		}
	}
}

void EnemyEntity::render( sf::RenderWindow& window )
{
	if ( health > 0 or type == "towerBoss" )
	{
		Entity::render( window );
	}
}

sf::FloatRect EnemyEntity::getCollisionRect()
{
	sf::FloatRect box;
	box.left = -( spr.getScale().x * spr.getOrigin().x );
	box.top = -( spr.getScale().y * spr.getOrigin().y );
	box.width = std::abs( box.left * 2 );
	box.height = std::abs( box.top * 2 );

	box.left += spr.getPosition().x - ( spr.getOrigin().x * spr.getScale().x ) + 1;
	box.top += spr.getPosition().y - ( spr.getOrigin().y * spr.getScale().y ) + 1;

	//log("%f %f vs %f %f\n",box.left+box.width/2,box.top+box.height/2,spr.getPosition().x,spr.getPosition().y );
	//log("%f %f vs %f %f\n\n",box.width,box.height,spr.getGlobalBounds().width,spr.getGlobalBounds().height);

	box = spr.getGlobalBounds();
	float oldW = box.width;
	float oldH = box.height;

	box.left = spr.getPosition().x - ( spr.getOrigin().x * spr.getScale().x * 0.9 ) + 0.5;
	box.top = spr.getPosition().y - ( spr.getOrigin().y * spr.getScale().y * 0.9 ) + 0.5;
	box.width = spr.getOrigin().x * 2 * spr.getScale().x * 0.9;
	box.height = spr.getOrigin().y * 2 * spr.getScale().y * 0.9;

	return box;
}

std::string EnemyEntity::getType() const
{
	return type;
}

int EnemyEntity::getHealth() const
{
	return health;
}

void EnemyEntity::setHealth( int theHealth )
{
	health = theHealth;
	startHealth = health;
}

void EnemyEntity::setStrength( int theStrength )
{
	strength = theStrength;
}

void EnemyEntity::setSpeed( float theSpeed )
{
	speed = theSpeed;
}

void EnemyEntity::setTexture( const sf::Texture& tex )
{
	spr.setTexture( tex, true );
	if ( type != "towerBoss" )
	{
		spr.setOrigin( tex.getSize().x / 2, tex.getSize().y / 2 );
	}
	else
	{
		spr.setOrigin( 8, 8 );
		spr.setTextureRect( sf::IntRect( 0, 0, 16, 16 ) );
	}
}

void EnemyEntity::hurt( int amt )
{
	if ( health > 0 )
	{
		health -= amt;

		player.playSound( "res/sound/hurt-enemy.wav" );

		if ( health <= 0 )
		{
			util::log( "Enemy \"%s\" dead.\n", type );
			if ( rand() % ( player.isHardMode() ? 11 : 9 ) == 0 )
			{
				util::log( "Dropping potion.\n" );
				room.addItem( "potion", spr.getPosition() );
			}

			if ( type == "dungeonBoss" )
			{
				util::log( "Dropping dungeon exit key.\n" );
				room.addItem( "key_dungeon3", spr.getPosition() );
				player.playSound( "res/sound/boss1-dead.wav" );
			}
			else if ( type == "towerBoss" )
			{
				player.playSound( "res/sound/boss2-dead.wav" );
				spr.setTextureRect( sf::IntRect( 8, 16, 8, 8 ) );
				spr.setScale( 1.f, 1.f );
				spr.setRotation( 0 );
				spr.setOrigin( 4, 4 );

				Room::Exit exit;
				exit.collision = sf::IntRect( 13, 13, 4, 4 );
				exit.area = "tower-top";
				exit.room = sf::Vector2i( 0, 0 );
				exit.pos = sf::Vector2i( 15, 9 );
				exit.giveHealth = false;

				room.getExits().push_back( exit );
			}
		}
		else if ( type.length() >= 4 and type.substr( type.length() - 4 ) != "Boss" )
		{
			stunTime = player.isHardMode() ? 20 : 60;
		}
	}
}

void EnemyEntity::reset()
{
	if ( health <= 0 ) return;

	spr.setPosition( sf::Vector2f( startPos ) );
	health = startHealth;
}
