#include "Player.hpp"

#include <cmath>
#include <SFML/Graphics/Texture.hpp>
#include <util/Logger.hpp>
#include <util/Math.hpp>

#include "Game.hpp"
#include "game/Area.hpp"
#include "game/BombEntity.hpp"
#include "game/EnemyEntity.hpp"
#include "game/Entity.hpp"
#include "game/ItemEntity.hpp"
#include "game/Room.hpp"
#include "scenes/GameScene.hpp"

Player::Player( Game& theGame, GameScene& theGameScene )
   : game( theGame ),
     scene( theGameScene ),

     up( theGame.options.getInputBinding( "Up" ) ),
     down( theGame.options.getInputBinding( "Down" ) ),
     left( theGame.options.getInputBinding( "Left" ) ),
     right( theGame.options.getInputBinding( "Right" ) ),
     action( theGame.options.getInputBinding( "Action" ) ),
     drop( theGame.options.getInputBinding( "Drop Item" ) ),
     slot1( theGame.options.getInputBinding( "Slot 1" ) ),
     slot2( theGame.options.getInputBinding( "Slot 2" ) ),
     slot3( theGame.options.getInputBinding( "Slot 3" ) ),
     slot4( theGame.options.getInputBinding( "Slot 4" ) ),
     slot5( theGame.options.getInputBinding( "Slot 5" ) ),
     prevItem( theGame.options.getInputBinding( "Prev Item" ) ),
     nextItem( theGame.options.getInputBinding( "Next Item" ) ),

     area( "surface" ),
     room( 0, 0 ),

     vel( 0.f, 0.f ),
     forceVel( 0.f, 0.f ),

     health( 12 ),
     healthMax( 12 ),
     invulnTime( -1 ),

     selected( 0 ),
     useTime( -1 ),
     startUseTime( -1 ),

     hasKitty( false ),
     kittyDidHeal( false )
{
	spr.setTexture( ( * game.getTexture( "res/entities/player.png" ) ), true );
	spr.setOrigin( 2, 2 );
	spr.setPosition( 15, 15 );
	spr.setScale( 0.35f, 0.35f );

	kittySpr.setTexture( * game.getTexture( "res/entities/kitty.png" ), true );
	kittySpr.setOrigin( 8, 16 );
	kittySpr.setPosition( spr.getPosition() );
	kittySpr.setScale( 0.05f, 0.05f );

	items.resize( 5 );
}

void Player::update()
{
	if ( area == "gameover" )
	{
		health = healthMax = 0;
	}

	// Movement
	sf::Vector2f movement( ( std::min( right.getState(), 0.7f ) - std::min( left.getState(), 0.7f ) ) / 0.7f,
	                       ( std::min(  down.getState(), 0.7f ) - std::min(   up.getState(), 0.7f ) ) / 0.7f );
	vel.x += movement.x * 0.32;
	vel.y += movement.y * 0.32;
	if ( std::abs( movement.x ) > std::abs( movement.y ) )
	{
		spr.setRotation( ( movement.x < 0 ) ? -90 : 90 );
	}
	else if ( std::abs( movement.x ) < std::abs( movement.y ) )
	{
		spr.setRotation( ( movement.y < 0 ) ? 0 : 180 );
	}

	     if ( vel.x >=  0.01f ) vel.x -= 0.01f;
	else if ( vel.x <= -0.01f ) vel.x += 0.01f;
	else                        vel.x  = 0.00f;

	     if ( vel.y >=  0.01f ) vel.y -= 0.01f;
	else if ( vel.y <= -0.01f ) vel.y += 0.01f;
	else                        vel.y  = 0.00f;

	     if ( forceVel.x >=  0.05f ) forceVel.x -= 0.05f;
	else if ( forceVel.x <= -0.05f ) forceVel.x += 0.05f;
	else                             forceVel.x  = 0.00f;

	     if ( forceVel.y >=  0.05f ) forceVel.y -= 0.05f;
	else if ( forceVel.y <= -0.05f ) forceVel.y += 0.05f;
	else                             forceVel.y  = 0.00f;

	vel += forceVel;

	Area& area = scene.getArea( this->area );
	Room& room = area.getRoom( this->room );

	// Exits
	for ( std::vector< Room::Exit >::iterator it = room.getExits().begin(); it != room.getExits().end(); ++it )
	{
		Room::Exit& exit = ( * it );
		sf::FloatRect coll( exit.collision );
		coll.left -= 0.1f;
		coll.top -= 0.1f;
		coll.width += 0.2f;
		coll.height += 0.2f;

		if ( coll.contains( spr.getPosition() ) )
		{
			if ( exit.itemReq == "" )//or hasItem( exit.itemReq ) )
			{
				this->area = exit.area;
				this->room = exit.room;
				spr.setPosition( sf::Vector2f( exit.pos ) );

				if ( exit.giveHealth )
				{
					//healthMax += 4;
					health = healthMax;
					exit.giveHealth = false;

					scene.playSound( "res/sound/teleport.wav" );
				}

				scene.getArea( this->area ).getRoom( this->room ).resetEnemies();

				return; // So we don't recollide on the other side
			}
		}
	}

	// Solids
	for ( std::vector< sf::IntRect >::const_iterator it = room.getSolids().begin(); it != room.getSolids().end(); ++it )
	{
		const sf::IntRect& solid = ( * it );
		if ( solid.contains( spr.getPosition().x + vel.x, spr.getPosition().y + vel.y ) )
		{
			// +X
			if ( spr.getPosition().x <= solid.left and spr.getPosition().x + vel.x > solid.left )
			{
				vel.x = solid.left - spr.getPosition().x;
			}

			// -X
			if ( spr.getPosition().x >= solid.left + solid.width and spr.getPosition().x + vel.x < solid.left + solid.width )
			{
				vel.x = ( solid.left + solid.width ) - spr.getPosition().x;
			}

			// +Y
			if ( spr.getPosition().y <= solid.top and spr.getPosition().y + vel.y > solid.top )
			{
				vel.y = solid.top - spr.getPosition().y;
			}

			// -Y
			if ( spr.getPosition().y >= solid.top + solid.height and spr.getPosition().y + vel.y < solid.top + solid.height )
			{
				vel.y = ( solid.top + solid.height ) - spr.getPosition().y;
			}
		}
	}

	spr.move( vel.x/2,vel.y/2 ); // Why did I suddenly have to do this? I have no clue
	vel = sf::Vector2f( 0.f, 0.f );

	// Boundaries
	if ( spr.getPosition().x < room.getBounds().left )
	{
		if ( room.canExitSides() and room.getPosition().x > area.getMinimumRoom().x )
		{
			this->room.x -= 1;
			spr.move( room.getBounds().width, 0 );
			scene.getArea( this->area ).getRoom( this->room ).resetEnemies();
		}
		else
		{
			spr.setPosition( room.getBounds().left, spr.getPosition().y );
		}
	}
	else if ( spr.getPosition().x > room.getBounds().left + room.getBounds().width )
	{
		if ( room.canExitSides() and room.getPosition().x < area.getMaximumRoom().x )
		{
			this->room.x += 1;
			spr.move( -room.getBounds().width, 0 );
			scene.getArea( this->area ).getRoom( this->room ).resetEnemies();
		}
		else
		{
			spr.setPosition( room.getBounds().left + room.getBounds().width, spr.getPosition().y );
		}
	}
	if ( spr.getPosition().y < room.getBounds().top )
	{
		if ( room.canExitSides() and room.getPosition().y > area.getMinimumRoom().y )
		{
			this->room.y -= 1;
			spr.move( 0, room.getBounds().height );
			scene.getArea( this->area ).getRoom( this->room ).resetEnemies();
		}
		else
		{
			spr.setPosition( spr.getPosition().x, room.getBounds().top );
		}
	}
	else if ( spr.getPosition().y > room.getBounds().top + room.getBounds().height )
	{
		if ( room.canExitSides() and room.getPosition().y < area.getMaximumRoom().y )
		{
			this->room.y += 1;
			spr.move( 0, -room.getBounds().height );
			scene.getArea( this->area ).getRoom( this->room ).resetEnemies();
		}
		else
		{
			spr.setPosition( spr.getPosition().x, room.getBounds().top + room.getBounds().height );
		}
	}

	// Interaction
	if ( action.justActive() )
	{
		sf::FloatRect bounds = spr.getGlobalBounds();
		bounds.left -= 1.f;
		bounds.top -= 1.f;
		bounds.width += 2.f;
		bounds.height += 2.f;

		// (Locked) Exits
		for ( std::vector< Room::Exit >::iterator it = room.getExits().begin(); it != room.getExits().end(); ++it )
		{
			Room::Exit& exit = ( * it );
			if ( sf::FloatRect( exit.collision ).intersects( bounds ) and exit.itemReq != "" )
			{
				if ( items[ selected ] == exit.itemReq )
				{
					scene.playSound( "res/sound/unlock.wav" );
					util::log( "Unlocked exit with %s.\n", items[ selected ] );
					items[ selected ] = "";
					exit.itemReq = "";
					return;
				}
			}
		}

		// Chests
		for ( std::vector< Room::Chest >::iterator it = room.getChests().begin(); it != room.getChests().end(); ++it )
		{
			const Room::Chest& chest = ( * it );
			if ( sf::FloatRect( chest.collision ).intersects( bounds ) )
			{
				std::size_t i;
				for ( i = 0; i < items.size(); ++i )
				{
					if ( items[ i ] == "" )
					{
						break;
					}
				}

				if ( i < items.size() )
				{
					util::log( "Obtained item %s.\n", chest.item );
					items[ i ] = chest.item;
					room.getChests().erase( it );
					scene.playSound( "res/sound/chest.wav" );
					return;
				}
			}
		}

		// Using an item
		if ( useTime < 0 and items[ selected ] != "" )
		{
			useItem( items[ selected ] );
		}
	}
	if ( useTime < 0 )
	{
		// Item selection
		if ( slot1.isActive() ) selected = 0;
		if ( slot2.isActive() ) selected = 1;
		if ( slot3.isActive() ) selected = 2;
		if ( slot4.isActive() ) selected = 3;
		if ( slot5.isActive() ) selected = 4;

		if ( prevItem.justActive() ) --selected;
		if ( nextItem.justActive() ) ++selected;
		     if ( selected < 0 ) selected = 4;
		else if ( selected > 4 ) selected = 0;

		// Item dropping
		if ( drop.justActive() and items[ selected ] != "" )
		{
			std::shared_ptr< ItemEntity > item = std::make_shared< ItemEntity > ( room,( * this ), items[ selected ], spr.getPosition() );
			item->setTexture( * game.getTexture( "res/items/" + items[ selected ] + ".png" ) );
			room.getEntities().push_back( std::static_pointer_cast< Entity >( item ) );
			items[ selected ] = "";
		}
	}

	// Timers
	if ( useTime >= 0 )
	{
		--useTime;
	}
	if ( invulnTime >= 0 )
	{
		--invulnTime;
	}

	// MEOW!
	if ( hasKitty )
	{
		if ( util::distance( kittySpr.getPosition(), spr.getPosition() ) > 10.f )
		{
			kittySpr.setPosition( spr.getPosition() );
		}

		float diffX = kittySpr.getPosition().x - spr.getPosition().x;
		float diffY = kittySpr.getPosition().y - spr.getPosition().y;
		float dirRad = std::atan2( diffX, diffY );
		float dirDeg = util::radiansToDegrees( dirRad );

		kittySpr.setRotation( -dirDeg );

		if ( util::distance( kittySpr.getPosition(), spr.getPosition() ) > 2.5f )
		{
			float offX = std::sin( dirRad ) * -0.25f;
			float offY = std::cos( dirRad ) * -0.25f;
			kittySpr.move( offX, offY );
		}
	}
}

void Player::render( sf::RenderWindow& window )
{
	if ( hasKitty )
	{
		window.draw( kittySpr );
	}

	if ( invulnTime % 20 < 10 )
	{
		window.draw( spr );
	}

	if ( useTime >= 0 )
	{
		sf::Texture& itemTex = ( * game.getTexture( "res/items/" + items[ selected ] + ".png" ) );

		sf::Sprite itemSpr = spr;
		itemSpr.setTexture( itemTex, true );
		itemSpr.setOrigin( 4, 12 );
		itemSpr.setScale( spr.getScale().x / 3 * 2, spr.getScale().y / 3 * 2 );

		window.draw( itemSpr );
	}
}

bool Player::isHardMode() const
{
	return scene.isHardMode();
}

const std::string& Player::getArea() const
{
	return area;
}

sf::Vector2i Player::getRoom() const
{
	return room;
}

sf::Vector2f Player::getPosition() const
{
	return spr.getPosition();
}

sf::FloatRect Player::getCollisionRect() const
{
	sf::FloatRect box;
	box.left = -( spr.getScale().x * ( spr.getTexture()->getSize().x / 2 ) ) * 0.9;
	box.top = -( spr.getScale().y * ( spr.getTexture()->getSize().x / 2 ) ) * 0.9;
	box.width = std::abs( box.left * 2 ) * 0.9;
	box.height = std::abs( box.top * 2 ) * 0.9;

	box.left += spr.getPosition().x + ( spr.getOrigin().x * spr.getScale().x );
	box.top += spr.getPosition().y + ( spr.getOrigin().x * spr.getScale().x );

	return box;
}

void Player::moveTo( const std::string& theArea, sf::Vector2i theRoom, sf::Vector2f pos )
{
	area = theArea;
	room = theRoom;
	spr.setPosition( pos );
}

int Player::getHealth() const
{
	return health;
}

int Player::getHealthMax() const
{
	return healthMax;
}

bool Player::hasPetKitty() const
{
	return hasKitty;
}

bool Player::didKittyHeal() const
{
	return kittyDidHeal;
}

void Player::hurt( int amt )
{
	if ( amt < 0 )
	{
		scene.playSound( "res/sound/heal.wav" );
		health = std::min( healthMax, health - amt );
	}
	else if ( invulnTime < 0 )
	{
		scene.playSound( "res/sound/hurt-player.wav" );
		health -= amt;
		invulnTime = scene.isHardMode() ? 35 : 50;
	}
}

void Player::kittyHeal()
{
	if ( hasKitty and !kittyDidHeal )
	{
		kittyDidHeal = true;
		health = healthMax;
		scene.playSound( "res/sound/heal.wav" );
	}
}

void Player::push( float force, float dir )
{
	float dirRad = dir;//util::degreesToRadians( spr.getRotation() );

	forceVel.x += std::sin( dirRad ) * force;
	forceVel.y += std::cos( dirRad ) * force;
}

bool Player::hasItem( const std::string& item ) const
{
	for ( std::size_t i = 0; i < items.size(); ++i )
	{
		if ( items[ i ] == item )
		{
			return true;
		}
	}

	return true;//false;
}

std::vector< std::string>& Player::getItems()
{
	return items;
}

const std::vector< std::string>& Player::getItems() const
{
	return items;
}

void Player::setSelectedItem( unsigned int theSelected )
{
	selected = theSelected;
}

unsigned int Player::getSelectedItem() const
{
	return selected;
}

int Player::getUseTime() const
{
	return useTime;
}

int Player::getStartingUseTime() const
{
	return startUseTime;
}

void Player::playSound( const std::string& path )
{
	scene.playSound( path );
}

void Player::useItem( const std::string& item )
{
	Area& area = scene.getArea( this->area );
	Room& room = area.getRoom( this->room );

	if ( item == "dagger" )
	{
		startUseTime = useTime = 15;
		scene.playSound( "res/sound/swipe.wav" );

		sf::Texture& itemTex = ( * game.getTexture( "res/items/" + items[ selected ] + ".png" ) );

		sf::Sprite itemSpr = spr;
		itemSpr.setTexture( itemTex, true );
		itemSpr.setOrigin( 4, 12 );
		itemSpr.setScale( spr.getScale().x / 3 * 2, spr.getScale().y / 3 * 2 );

		sf::FloatRect rect = itemSpr.getGlobalBounds();
		for ( std::vector< std::shared_ptr< Entity > >::iterator it = room.getEntities().begin(); it != room.getEntities().end(); ++it )
		{
			Entity* entity = it->get();
			if ( EnemyEntity* enemy = dynamic_cast< EnemyEntity* >( entity ) )
			{
				if ( enemy->getCollisionRect().intersects( rect ) and enemy->getHealth() > 0 )
				{
					enemy->hurt( 2 );
					break;
				}
			}
		}
	}
	else if ( item == "leaf" )
	{
		startUseTime = useTime = 75;
		scene.playSound( "res/sound/blow.wav" ); //

		sf::Texture& itemTex = ( * game.getTexture( "res/items/" + items[ selected ] + ".png" ) );

		sf::Sprite itemSpr = spr;
		itemSpr.setTexture( itemTex, true );
		itemSpr.setOrigin( 4, 8 );
		itemSpr.setScale( 0.75f, 3.5f );

		sf::FloatRect bigRect = itemSpr.getGlobalBounds();

		itemSpr.setScale( spr.getScale().x / 3 * 2, spr.getScale().y / 3 * 2 );
		itemSpr.scale( 1.25f, 1.25f );

		sf::FloatRect smallRect = itemSpr.getGlobalBounds();
		for ( std::vector< std::shared_ptr< Entity > >::iterator it = room.getEntities().begin(); it != room.getEntities().end(); ++it )
		{
			Entity* entity = it->get();

			if ( entity->getCollisionRect().intersects( bigRect ) )
			{
				float diffX = entity->getPosition().x - spr.getPosition().x;
				float diffY = entity->getPosition().y - spr.getPosition().y;
				float dirRad = std::atan2( diffX, diffY );

				entity->push( 0.5f, dirRad );
			}
		}

		for ( std::vector< std::shared_ptr< Entity > >::iterator it = room.getEntities().begin(); it != room.getEntities().end(); ++it )
		{
			Entity* entity = it->get();

			if ( entity->getCollisionRect().intersects( smallRect ) )
			{
				float diffX = entity->getPosition().x - spr.getPosition().x;
				float diffY = entity->getPosition().y - spr.getPosition().y;
				float dirRad = std::atan2( diffX, diffY );

				entity->push( 0.3f, dirRad );
				if ( EnemyEntity* enemy = dynamic_cast< EnemyEntity* >( entity ) )
				{
					if ( enemy->getHealth() > 0 )
					{
						enemy->hurt( 1 );
					}
				}
			}
		}
	}
	else if ( item == "potion" )
	{
		hurt( scene.isHardMode() ? -4 : ( health - healthMax ) );
		items[ selected ] = "";
	}
	else if ( item == "kitty" )
	{
		items[ selected ] = "";
		hasKitty = true;
	}
	else if ( item == "bomb" )
	{
		startUseTime = useTime = 75;
		scene.playSound( "res/sound/drop.wav" );

		std::shared_ptr< BombEntity > bomb = std::make_shared< BombEntity >( room, ( * this ), spr.getPosition() );
		bomb->setTexture( * game.getTexture( "res/entities/bomb.png" ) );
		room.getEntities().push_back( bomb );
	}
	else if ( item == "heart" )
	{
		items[ selected ] = "";
		healthMax += 4;
		health = healthMax;
		scene.playSound( "res/sound/health-max.wav" );
	}
	else if ( item == "potato" )
	{
		//startUseTime = useTime = 250;
		scene.playSound( "res/sound/swipe.wav" );

		sf::Texture& itemTex = ( * game.getTexture( "res/items/" + items[ selected ] + ".png" ) );

		sf::Sprite itemSpr = spr;
		itemSpr.setTexture( itemTex, true );
		itemSpr.setOrigin( 4, 4 );
		itemSpr.setScale( 100, 100 );

		sf::FloatRect rect = itemSpr.getGlobalBounds();
		for ( std::vector< std::shared_ptr< Entity > >::iterator it = room.getEntities().begin(); it != room.getEntities().end(); ++it )
		{
			Entity* entity = it->get();
			if ( !entity )
			{
				continue;
			}

			if ( EnemyEntity* enemy = dynamic_cast< EnemyEntity* >( entity ) )
			{
				if ( enemy->getCollisionRect().intersects( rect ) and enemy->getHealth() > 0 )
				{
					enemy->hurt( 64 );
				}
			}
		}
	}
}
