#include "scenes/GameScene.hpp"

#include <ctime>
#include <iostream>
#include <fstream>
#include <SFML/Graphics.hpp>
#include <util/Logger.hpp>

#include "Game.hpp"
#include "game/EnemyEntity.hpp"
#include "LoaderUtils.hpp"
#include "scenes/MainMenuScene.hpp"

GameScene::GameScene( Game& theGame, const SceneChangeEvent& event )
   : Scene::Scene( theGame ),
     player( theGame, ( * this ) ),
     paused( false ),
     hard( ( event.reason == SceneChangeEvent::PlayHard ) ? true : false )
{
	loadWorld();
	music.setVolume( 10.f );
}

void GameScene::update()
{
	game.getWindow().setView( sf::View( sf::FloatRect( 0, 0, Game::TILE_AMOUNT, Game::TILE_AMOUNT ) ) );

	if ( game.options.getInputBinding( "Pause" ).justActive() )
	{
		if ( player.getArea() == "gameover" or player.getArea() == "tower-top" )
		{
			game.changeScenes< MainMenuScene >();
		}
		else
		{
			paused = !paused;
		}
	}

	if ( !paused )
	{
		// Entities
		Room& room = areas.at( player.getArea() )->getRoom( player.getRoom() );
		for ( std::size_t i = 0; i < room.getEntities().size(); ++i )
		{
			room.getEntities()[ i ]->update();
		}

		// Player
		player.update();

		// Gameover
		if ( player.getHealth() <= 0 and player.getHealthMax() > 0 )
		{
			player.kittyHeal();

			if ( player.getHealth() <= 0 and player.getHealthMax() > 0 )
			{
				player.moveTo( "gameover", sf::Vector2i( 0, 0 ), sf::Vector2f( 15, 15 ) );
			}
		}
	}

	// Sounds
	for ( std::list< sf::Sound >::iterator it = sounds.begin(); it != sounds.end(); ++it )
	{
		if ( it->getStatus() != sf::Sound::Playing )
		{
			std::list< sf::Sound >::iterator it2 = it; ++it2;
			sounds.erase( it );
			it = it2;
		}
	}

	// Music
	if ( currMusic != getTargetMusic() )
	{
		if ( music.getVolume() > 1.f )
		{
			music.setVolume( music.getVolume() - 0.25f );
		}
		else
		{
			currMusic = getTargetMusic();
			music.openFromFile( "res/music/" + getTargetMusic() + ".ogg" );
			music.setLoop( true );
			music.play();
		}
	}
	else
	{
		if ( music.getVolume() < 30.f )
		{
			music.setVolume( music.getVolume() + 0.25f );
		}
	}

	// Hard mode unlock
	if ( !isHardMode() and player.getArea() == "tower-top" and !game.options.hardUnlocked )
	{
		game.options.hardUnlocked = true;
	}
}

void GameScene::update( const sf::Event& event )
{
}

void GameScene::render( sf::RenderWindow& window )
{
	// Background
	std::string texPath = "res/" + player.getArea() + "/" + util::toString( player.getRoom().x ) + "_" + util::toString( player.getRoom().y ) + ".png";
	sf::Texture& tex = ( * game.getTexture( texPath ) );
	sf::Sprite spr( tex );
	window.draw( spr );

	// Entities
	Room& room = areas.at( player.getArea() )->getRoom( player.getRoom() );
	for ( std::vector< std::shared_ptr< Entity > >::iterator it = room.getEntities().begin(); it != room.getEntities().end(); ++it )
	{
		( * it )->render( window );
	}

	// Player
	player.render( window );

	if ( player.getHealthMax() <= 0 )
	{
		return;
	}

	// Item hotbar
	sf::Texture& slotTex = ( * game.getTexture( "res/gui/slot.png" ) );
	sf::Texture& selectedTex = ( * game.getTexture( "res/gui/selected.png" ) );
	for ( std::size_t i = 0; i < player.getItems().size(); ++i )
	{
		const std::string& item = player.getItems()[ i ];

		sf::Sprite slotSpr( ( player.getSelectedItem() == i ) ? selectedTex : slotTex );
		slotSpr.setPosition( 6.75f + ( i * 2.5f ) + ( i * 1 ), 30 - 2.5f - 1 );
		slotSpr.setScale( 0.25f, 0.25f );

		window.draw( slotSpr );

		if ( item != "" )
		{
			sf::Texture& itemTex = ( * game.getTexture( "res/items/" + item + ".png" ) );

			sf::Sprite itemSpr( itemTex );
			itemSpr.setPosition( slotSpr.getPosition().x + 0.25, slotSpr.getPosition().y + 0.25 );
			itemSpr.setScale( 0.25f, 0.25f );

			if ( item == "kitty" )
			{
				itemSpr.setScale( 0.125f, 0.125f );
			}

			window.draw( itemSpr );
		}

		if ( i == player.getSelectedItem() )
		{
			if ( player.getItems()[ i ] != "" )
			{
				sf::Text text( player.getItems()[ i ], ( * game.getFont( "res/ken_fonts/kenpixel.ttf" ) ), 30 );
				if ( std::string( text.getString() ).substr( 0, 4 ) == "key_" )
				{
					text.setString( std::string( text.getString() ).substr( 4 ) + " Key" );
				}
				text.setColor( sf::Color::White );
				text.setPosition( Game::TILE_AMOUNT / 2, slotSpr.getPosition().y - 1 );
				text.setOrigin( text.getLocalBounds().width / 2, text.getGlobalBounds().height / 2 );
				text.setScale( Game::TILE_SIZE / static_cast< float >( Game::WINDOW_SIZE ), Game::TILE_SIZE / static_cast< float >( Game::WINDOW_SIZE ) );
				window.draw( text );
			}

			if ( player.getUseTime() >= 0 )
			{
				float percent = static_cast< float >( player.getUseTime() ) / player.getStartingUseTime();
				int frame = std::min< int >( 8 * percent, 7 );

				sf::Texture& timerTex = ( * game.getTexture( "res/gui/timer.png" ) );

				sf::Sprite timerSpr( timerTex );
				timerSpr.setTextureRect( sf::IntRect( 8 * frame, 0, 8, 8 ) );
				timerSpr.setPosition( slotSpr.getPosition().x + 0.25, slotSpr.getPosition().y + 0.25 );
				timerSpr.setScale( 0.25f, 0.25f );
				timerSpr.setColor( sf::Color( 255, 255, 255, 128 ) );

				window.draw( timerSpr );
			}
		}
	}

	// Health
	int hearts = ( player.getHealthMax() / 4.f ) + 0.5;
	int startX = ( 30 - ( ( hearts * 2.75f ) + ( ( hearts - 1 ) * 1.25f ) ) ) / 2;
	sf::Texture& healthTex = ( * game.getTexture( "res/gui/health.png" ) );
	for ( int i = 0; i < hearts; ++i )
	{
		int frame = 0;
		if ( ( i + 1 ) * 4 <= player.getHealth() )
		{
			frame = 4;
		}
		else if ( i * 4 > player.getHealth() )
		{
			frame = 0;
		}
		else
		{
			frame = player.getHealth() % 4;
		}

		sf::Sprite slotSpr( healthTex );
		slotSpr.setPosition( startX + ( i * 2.75f ) + ( i * 1.25f ), 1.f );
		slotSpr.setScale( 0.5f, 0.5f );
		slotSpr.setTextureRect( sf::IntRect( frame * 8, 0, 8, 8 ) );

		window.draw( slotSpr );
	}

	if ( player.hasPetKitty() and !player.didKittyHeal() )
	{
		sf::Sprite catSpr( * game.getTexture( "res/items/kitty.png" ) );
		catSpr.setPosition( startX - 2.75f - 1.25f, 1.f );
		catSpr.setScale( 0.25f, 0.25f );
		window.draw( catSpr );
	}

	if ( paused )
	{
		sf::RectangleShape rect;
		rect.setSize( window.getView().getSize() );
		rect.setFillColor( sf::Color( 0, 0, 0, 128 ) );
		window.draw( rect );

		sf::Text text( "Paused", ( * game.getFont( "res/ken_fonts/kenpixel.ttf" ) ), 30 );
		text.setColor( sf::Color::White );
		text.setPosition( rect.getSize().x / 2, rect.getSize().y / 2 );
		text.setOrigin( text.getLocalBounds().width / 2, text.getGlobalBounds().height / 2 );
		text.setScale( rect.getSize().x / window.getSize().x, rect.getSize().y / window.getSize().y );
		window.draw( text );
	}
}

Area& GameScene::getArea( const std::string& name )
{
	return ( * areas.at( name ) );
}

const Area& GameScene::getArea( const std::string& name ) const
{
	return ( * areas.at( name ) );
}

Player& GameScene::getPlayer()
{
	return player;
}

const Player& GameScene::getPlayer() const
{
	return player;
}

void GameScene::playSound( const std::string& path )
{
	if ( loadingWorld )
	{
		return;
	}

	sounds.push_back( sf::Sound() );
	sounds.back().setBuffer( * game.getSoundBuffer( path ) );
	if ( path == "res/sound/blow.wav" or path == "res/sound/blow.wav" )
	{
		//sounds.back().setVolume( 1.f );
	}
	sounds.back().play();
}

bool GameScene::isHardMode() const
{
	return hard;
}

void GameScene::loadWorld()
{
	loadingWorld = true;

	util::log( "Creating world...\n" );
	std::ifstream file( "res/world.txt", std::ifstream::in );
	while ( file )
	{
		std::string line;
		std::getline( file, line );
		if ( file and line.length() >= 6 )
		{
			std::pair< std::string, std::string > parts = LoaderUtils::parseSides( line );
			std::string name = parts.first;
			sf::Vector2i size = LoaderUtils::parseVec2( parts.second ).data;

			std::shared_ptr< Area > area = std::make_shared< Area >( game, player, name, size );
			areas.insert( std::make_pair( name, area ) );
		}
	}

	player.moveTo( "surface", sf::Vector2i( 0, 0 ), sf::Vector2f( 15, 15 ) );

	loadingWorld = false;
}

std::string GameScene::getTargetMusic() const
{
	std::string music = player.getArea();

	bool boss = false;
	Room& room = areas.at( player.getArea() )->getRoom( player.getRoom() );
	for ( std::vector< std::shared_ptr< Entity > >::iterator it = room.getEntities().begin(); it != room.getEntities().end(); ++it )
	{
		Entity* entity = it->get();
		if ( EnemyEntity* enemy = dynamic_cast< EnemyEntity* >( entity ) )
		{
			if ( enemy->getType().length() > 4 and enemy->getType().substr( enemy->getType().length() - 4 ) == "Boss" )
			{
				boss = true;
			}
		}
	}

	if ( boss )
	{
		music += "-boss";
	}

	return music;
}
